#! /usr/bin/env python

"""
Compute difference between two rasters
"""

import sys
import os
import re
import argparse
from datetime import timedelta

import numpy as np

from pygeotools.lib import timelib, iolib, malib, warplib

def getparser():
    parser = argparse.ArgumentParser(description="Compute difference between two rasters")
    parser.add_argument('-outdir', default=None, help='Output directory')
    parser.add_argument('-tr', default='max', help='Output resolution (default: %(default)s)')
    parser.add_argument('-te', default='intersection', help='Output extent (default: %(default)s)')
    parser.add_argument('-t_srs', default='first', help='Output projection (default: %(default)s)')
    parser.add_argument('-rate', action='store_true', help='Attempt to generate change rate products (dz/dt) in m/yr (default: %(default)s)')
    parser.add_argument('fn1', type=str, help='Raster filename 1')
    parser.add_argument('fn2', type=str, help='Raster filename 2')
    return parser

def main():
    parser = getparser()
    args = parser.parse_args()

    #This is output ndv, avoid using 0 for differences
    diffndv = -9999

    r1_fn = args.fn1
    r2_fn = args.fn2

    if r1_fn == r2_fn:
        sys.exit('Input filenames are identical')

    fn_list = [r1_fn, r2_fn]

    outdir = args.outdir
    if outdir is None:
        outdir = os.path.dirname(os.path.abspath(r1_fn))

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    outprefix = os.path.splitext(os.path.split(r1_fn)[1])[0]+'_'+os.path.splitext(os.path.split(r2_fn)[1])[0]

    #Compute dz/dt rate if possible, in m/yr
    if args.rate:
        #Extract basename
        #This was a hack to work with timestamp array filenames that have geoid offset applied
        adj = ''
        if '-adj' in r1_fn:
            adj = '-adj' 
        r1_fn_base = re.sub(adj, '', os.path.splitext(r1_fn)[0]) 
        r2_fn_base = re.sub(adj, '', os.path.splitext(r2_fn)[0]) 

        #Attempt to load ordinal timestamp arrays (for mosaics) if present
        """
        import glob
        t1_fn = glob.glob(r1_fn_base+'*_ts*.tif')
        t2_fn = glob.glob(r1_fn_base+'*_ts*.tif')
        t_unit = 'year'
        """
        t1_fn = r1_fn_base+'_ts.tif'
        t2_fn = r2_fn_base+'_ts.tif'
        t_unit = 'day'
        if not os.path.exists(t1_fn) and not os.path.exists(t2_fn):
            #Try to find processed output from r_mosaic index
            #These are decimal years
            t1_fn = r1_fn_base+'index_ts.tif'
            t2_fn = r2_fn_base+'index_ts.tif'
            t_unit = 'year'
        print(t1_fn, t2_fn)
        if os.path.exists(t1_fn) and os.path.exists(t2_fn):
            fn_list.extend([t1_fn, t2_fn])
        else:
            #Attempt to extract timestamps from input filenames
            t1 = timelib.fn_getdatetime(r1_fn)
            t2 = timelib.fn_getdatetime(r2_fn)
            if t1 is not None and t2 is not None and t1 != t2:  
                dt = t2 - t1
                year = timedelta(days=365.25)
                t_factor = abs(dt.total_seconds()/year.total_seconds()) 
                print("Time differences is %s, dh/%0.3f" % (dt, t_factor))
            else:
                print("Unable to extract timestamps for input images")
                args.rate = False


    print("Warping rasters to same res/extent/proj")
    #This will check input param for validity, could do beforehand
    ds_list = warplib.memwarp_multi_fn(fn_list, extent=args.te, res=args.tr, t_srs=args.t_srs, r='cubic')
    r1_ds = ds_list[0]
    r2_ds = ds_list[1]

    print("Loading input rasters into masked arrays")
    r1 = iolib.ds_getma(r1_ds, 1)
    r2 = iolib.ds_getma(r2_ds, 1)

    #Compute relative difference 
    print("Computing raster difference")
    diff = r2 - r1

    #Check to make sure inputs actually intersect
    if diff.count() == 0:
        sys.exit("No valid overlap between input rasters")

    if len(fn_list) == 4:
        t1_ds = ds_list[2]
        t2_ds = ds_list[3]
        print("Loading timestamps into masked arrays")
        t1 = iolib.ds_getma(t1_ds)
        t2 = iolib.ds_getma(t2_ds)
        #Compute dt in years 
        t_factor = t2 - t1
        if t_unit == 'day':
            t_factor /= 365.25

    if True:
        print("Raster difference stats:")
        diff_stats = malib.print_stats(diff)
        diff_med = diff_stats[5]

    if True:
        print("Writing raster difference map")
        dst_fn = os.path.join(outdir, outprefix+'_diff.tif')
        print(dst_fn)
        iolib.writeGTiff(diff, dst_fn, r1_ds, ndv=diffndv)
        if args.rate:
            print("Writing rate map")
            dst_fn = os.path.join(outdir, outprefix+'_diff_rate.tif')
            print(dst_fn)
            iolib.writeGTiff(diff/t_factor, dst_fn, r1_ds, ndv=diffndv)
            if len(fn_list) == 4:
                print("Writing time difference map")
                dst_fn = os.path.join(outdir, outprefix+'_diff_dt.tif')
                print(dst_fn)
                iolib.writeGTiff(t_factor, dst_fn, r1_ds, ndv=diffndv)

    if False:
        print("Writing relative raster difference map")
        diff_rel = diff - diff_med
        dst_fn = os.path.join(outdir, outprefix+'_diff_rel.tif')
        print(dst_fn)
        iolib.writeGTiff(diff_rel, dst_fn, r1_ds, ndv=diffndv)

    if False:
        print("Writing out raster2 with median difference removed")
        dst_fn = os.path.splitext(r2_fn)[0]+'_med'+diff_med+'.tif'
        print(dst_fn)
        iolib.writeGTiff(r2 - diff_med, dst_fn, r1_ds, ndv=diffndv)

    if False:
        print("Writing raster difference percentage map (relative to raster1)")
        diff_perc = 100.0*diff/r1
        dst_fn = os.path.join(outdir, outprefix+'_diff_perc.tif')
        print(dst_fn)
        iolib.writeGTiff(diff_perc, dst_fn, r1_ds, ndv=diffndv)

if __name__ == "__main__":
    main()



#! /usr/bin/env python

#David Shean
#dshean@gmail.com

#Take translation-only transform from ASP pc_align, update projected geotransform and apply z offset to DEM
#This offers significant performance gains over writing out new PC and running point2dem

import sys, os
import argparse

import numpy as np
from osgeo import gdal, osr

from pygeotools.lib import iolib

#This should probably be moved to geolib
def get_proj_shift(src_c, src_shift, s_srs, t_srs, inv_trans=False):
    if s_srs.IsSame(t_srs):
        proj_shift = src_shift
    else:
        src_c_shift = src_c + src_shift
        src2proj = osr.CoordinateTransformation(s_srs, t_srs)
        proj_c = np.array(src2proj.TransformPoint(*src_c))
        proj_c_shift = np.array(src2proj.TransformPoint(*src_c_shift))
        if inv_trans:
            proj_shift = proj_c - proj_c_shift
        else:
            proj_shift = proj_c_shift - proj_c
    #Reduce unnecessary precision
    proj_shift = np.around(proj_shift, 3)
    return proj_shift

def getparser():
    parser = argparse.ArgumentParser(description="Apply existing pc_align translation to a DEM")
    parser.add_argument('dem_fn', type=str, help='DEM filename')
    parser.add_argument('log_fn', type=str, help='pc_align log filename')
    parser.add_argument('-outdir', default=None, help='Output directory')
    return parser

def main():
    parser = getparser()
    args = parser.parse_args()
  
    dem_fn = args.dem_fn
    log_fn = args.log_fn
    outdir = args.outdir

    if not iolib.fn_check(dem_fn): 
        sys.exit("Unable to find input DEM: %s" % dem_fn)
    if not iolib.fn_check(log_fn):
        sys.exit("Unable to find input log: %s" % log_fn)
    if outdir is not None:
        if not iolib.fn_check(outdir):
            os.makedirs(outdir)
    else:
        outdir = os.path.dirname(dem_fn)

    #Prepare input image
    src_ds = gdal.Open(dem_fn)
    print("Loading input DEM")
    src_a = iolib.ds_getma(src_ds)
    #Overhead for np.ma may be unnecessary here, could just load as np.array
    #src_a = src_ds.GetRasterBand(1).ReadAsArray()
    src_ndv = src_a.fill_value 

    #Extract info from input log
    #These are coordinates of source points
    ecef_centroid_str = 'Centroid of source points (Cartesian, meters):'
    llz_centroid_str = 'Centroid of source points (lat,lon,z):'
    ecef_shift_str = 'Translation vector (Cartesian, meters):'
    ned_shift_str = 'Translation vector (North-East-Down, meters):'
    llz_shift_str = 'Translation vector (lat,lon,z):'

    #Need to extract from log to know how to compute translation
    #if ref is csv and src is dem, want to transform source_center + shift
    inv_trans = False 
    #if ref is dem and src is csv, want to inverse transform ref by shift applied at (source_center - shift)
    #inv_trans_str = '--save-inv-transformed-reference-points'
    inv_trans_str = 'trans_reference: true'

    llz_c = None
    log = open(log_fn)
    for line in log:
        if inv_trans_str in line:
            inv_trans = True
        if ecef_centroid_str in line:
            ecef_c = np.fromstring(line.split('Vector3')[1][1:-1], sep=',')
        if llz_centroid_str in line:
            llz_c = np.fromstring(line.split('Vector3')[1][1:-1], sep=',')
        if ecef_shift_str in line:
            ecef_shift = np.fromstring(line.split('Vector3')[1][1:-1], sep=',')
        if ned_shift_str in line:
            ned_shift = np.fromstring(line.split('Vector3')[1][1:-1], sep=',')
        if llz_shift_str in line:
            llz_shift = np.fromstring(line.split('Vector3')[1][1:-1], sep=',')
            break
    log.close()

    if llz_c is None:
        sys.exit("Log file does not contain necessary translation information: %s" % log_fn)

    #Reorder lat,lon,z to lon,lat,z (x,y,z)
    i = [1, 0, 2]
    llz_c = llz_c[i]
    llz_shift = llz_shift[i]

    ecef_srs=osr.SpatialReference()
    ecef_srs.ImportFromEPSG(4978)

    s_srs = ecef_srs
    src_c = ecef_c
    src_shift = ecef_shift

    #Determine shift in original dataset coords
    t_srs = osr.SpatialReference()
    t_srs.ImportFromWkt(src_ds.GetProjectionRef())
    if t_srs is None:
        sys.exit("Unable to determine src_ds proj")
    proj_shift = get_proj_shift(src_c, src_shift, s_srs, t_srs, inv_trans)

    #Should compute distances in local projection
    #Local orthographic
    ortho_proj4 = '+proj=ortho +ellps=WGS84 +lat_0=%f +lon_0=%f' % (llz_c[1], llz_c[0])
    ortho_srs = osr.SpatialReference()
    ortho_srs.ImportFromProj4(ortho_proj4)
    ortho_shift = get_proj_shift(src_c, src_shift, s_srs, ortho_srs, inv_trans)

    #Local stereographic
    stereo_proj4 = '+proj=stere +ellps=WGS84 +lat_0=%f +lon_0=%f' % (llz_c[1], llz_c[0])
    stereo_srs = osr.SpatialReference()
    stereo_srs.ImportFromProj4(stereo_proj4)
    stereo_shift = get_proj_shift(src_c, src_shift, s_srs, stereo_srs, inv_trans)

    print("ECEF shift", ecef_shift, np.linalg.norm(ecef_shift))
    print("local ned shift", ned_shift, np.linalg.norm(ecef_shift))
    print("local ortho shift", ortho_shift, np.linalg.norm(ortho_shift))
    print("local stereo shift", stereo_shift, np.linalg.norm(stereo_shift))
    print("proj shift", proj_shift, np.linalg.norm(proj_shift))

    #out_fmt = "VRT"
    out_fmt = "TIF"
    out_fn = os.path.join(outdir, os.path.split(os.path.splitext(dem_fn)[0])[1] + '_trans')
    if out_fmt == "VRT": 
        print("Writing vrt with scaled values")
        dst_fn = out_fn+'.vrt'
        dst_ds = iolib.vrt_drv.CreateCopy(dst_fn, src_ds, 0)
        dst_ds.GetRasterBand(1).SetOffset(proj_shift[2])
    else:
        dst_fn = out_fn+'.tif'
        #Create might be faster here 
        print("Copying input dataset")
        dst_ds = iolib.gtif_drv.CreateCopy(dst_fn, src_ds, 0, options=iolib.gdal_opt)
        #Apply vertical shift
        dst_b = dst_ds.GetRasterBand(1)
        print("Writing out z-shifted band")
        dst_b.SetNoDataValue(float(src_ndv))
        dst_b.WriteArray(np.around((src_a + proj_shift[2]).filled(src_ndv), decimals=3))

    dst_gt = list(dst_ds.GetGeoTransform())
    #Apply horizontal shift directly to geotransform
    print("Updating geotransform with horizontal shift")
    dst_gt[0] += proj_shift[0]
    dst_gt[3] += proj_shift[1] 
    dst_ds.SetGeoTransform(dst_gt)

    print("Writing out: ", dst_fn)
    dst_ds = None
    src_ds = None

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from distutils.core import setup

#To prepare a new release
#python setup.py sdist upload

setup(name='demcoreg',
    version='0.5.0',
    description='Utilities for DEM co-registration',
    author='David Shean',
    author_email='dshean@gmail.com',
    license='MIT',
    url='https://github.com/dshean/demcoreg',
    packages=['demcoreg'],
    long_description=open('README.md').read(),
    install_requires=['numpy','gdal','pygeotools','wget'],
    #Note: this will create local copy of executable scripts
    #scripts=['demcoreg/pc_align_wrapper.sh', 'demcoreg/apply_dem_translation.py', 'demcoreg/compute_diff.py', 'demcoreg/dem_align.py', 'demcoreg/dem_mask.py', 'demcoreg/dem_coreg.sh', 'demcoreg/dem_coreg_all.sh', 'demcoreg/vol_stats.py', 'demcoreg/robust_stats.py', 'demcoreg/glas_proc.py', 'demcoreg/filter_glas.py', 'demcoreg/get_nlcd.sh', 'demcoreg/get_bareground.sh', 'demcoreg/get_rgi.sh']
)




#! /usr/bin/env python

#David Shean
#dshean@gmail.com

#Utility to process ICESat-1 GLAS products, filter and clip to specified bounding box
#Input is HDF5 GLAH14
#https://nsidc.org/data/GLAH14/versions/34
#http://nsidc.org/data/docs/daac/glas_altimetry/data-dictionary-glah14.html

import os, sys
from datetime import datetime, timedelta
import argparse

import h5py
import numpy as np
from osgeo import gdal

from pygeotools.lib import timelib, geolib, iolib, malib, filtlib

#This is needed for LULC products
import dem_mask

#Before running, download all GLAH14 products
#lftp ftp://n5eil01u.ecs.nsidc.org/DP5/GLAS/
#mirror --parallel=16 GLAH14.034

"""
cd GLAH14.034
lfs setstripe -c 32 .
for site in conus hma
do
    parallel --progress --delay 1 -j 32 "~/src/demcoreg/demcoreg/glas_proc.py {} $site" ::: */*.H5
    #Combine output
    for ext in ${site}.csv ${site}_refdemfilt.csv ${site}_refdemfilt_lulcfilt.csv
    do
        first=$(ls */*$ext | head -1)
        head -1 $first > GLAH14_$ext
        cat */*$ext | sort -n | grep -v lat >> GLAH14_$ext
    done
done
"""

#Clip to glacier polygons
#clipsrc=/Volumes/d/hma/rgi/rgi_hma_aea_110kmbuffer_wgs84.shp
#vrt=GLAH14_tllz_hma_lulcfilt_demfilt.vrt
#ogr2ogr -progress -overwrite -clipsrc $clipsrc ${vrt%.*}_clip.shp $vrt

def getparser():
    parser = argparse.ArgumentParser(description="Process and filter ICESat GLAS points")
    parser.add_argument('fn', type=str, help='GLAH14 HDF5 filename')
    site_choices = geolib.site_dict.keys()
    parser.add_argument('sitename', type=str, choices=site_choices, help='Site name')
    #parser.add_argument('--rockfilter', action='store_true', help='Only output points over exposed rock using NLCD or bareground')
    parser.add_argument('-extent', type=str, default=None, help='Specify output spatial extent ("xmin xmax ymin ymax"). Otherwise, use default specified for sitename in pygeotools/lib/geolib')
    parser.add_argument('-refdem_fn', type=str, default=None, help='Specify alternative reference DEM for filtering. Otherwise use NED or SRTM')
    return parser

def main():
    parser = getparser()
    args = parser.parse_args()

    fn = args.fn
    sitename = args.sitename
    #User-specified output extent
    #Note: not checked, untested
    if args.extent is not None:
        extent = (args.extent).split()
    else:
        extent = (geolib.site_dict[sitename]).extent
    if args.refdem_fn is not None:
        refdem_fn = args.refdem_fn
    else:
        refdem_fn = (geolib.site_dict[sitename]).refdem_fn
    
    #Max elevation difference between shot and sampled DEM
    max_z_DEM_diff = 200
    #Max elevation std for sampled DEM values in padded window around shot
    max_DEMhiresArElv_std = 50.0

    f = h5py.File(fn)
    t = f.get('Data_40HZ/Time/d_UTCTime_40')[:]

    #pyt0 = datetime(1, 1, 1, 0, 0)
    #utct0 = datetime(1970, 1, 1, 0, 0)
    #t0 = datetime(2000, 1, 1, 12, 0, 0)
    #offset_s = (t0 - utct0).total_seconds()
    offset_s = 946728000.0
    t += offset_s
    dt = timelib.np_utc2dt(t)
    dt_o = timelib.dt2o(dt)
    #dts = timelib.np_print_dt(dt)
    #dt_decyear = timelib.np_dt2decyear(dt)
    dt_int = np.array([ts.strftime('%Y%m%d') for ts in dt], dtype=long)

    lat = np.ma.masked_equal(f.get('Data_40HZ/Geolocation/d_lat')[:], 1.7976931348623157e+308)
    lon = np.ma.masked_equal(f.get('Data_40HZ/Geolocation/d_lon')[:], 1.7976931348623157e+308)
    lon = geolib.lon360to180(lon)
    z = np.ma.masked_equal(f.get('Data_40HZ/Elevation_Surfaces/d_elev')[:], 1.7976931348623157e+308)

    print('Input: %i' % z.count())

    #Now spatial filter - should do this up front
    x = lon
    y = lat
    xmin, xmax, ymin, ymax = extent
    #This is True if point is within extent
    valid_idx = ((x >= xmin) & (x <= xmax) & (y >= ymin) & (y <= ymax))

    #Prepare output array
    #out = np.ma.vstack([dt_decyear, dt_o, dt_int, lat, lon, z]).T
    out = np.ma.vstack([dt_o, dt_int, lat, lon, z]).T
    #Create a mask to ensure all four values are valid for each point
    mask = ~(np.any(np.ma.getmaskarray(out), axis=1))
    mask *= valid_idx
    out = out[mask]
    valid_idx = ~(np.any(np.ma.getmaskarray(out), axis=1))

    #Lon and lat indices
    xcol = 3
    ycol = 2
    zcol = 4

    if out.shape[0] == 0:
        sys.exit("No points within specified extent\n")
    else:
        print("Spatial filter: %i" % out.shape[0])

    #out_fmt = ['%0.8f', '%0.8f', '%i', '%0.6f', '%0.6f', '%0.2f'] 
    #out_hdr = ['dt_decyear, dt_ordinal', 'dt_YYYYMMDD', 'lat', 'lon', 'z_WGS84']
    out_fmt = ['%0.8f', '%i', '%0.6f', '%0.6f', '%0.2f'] 
    out_hdr = ['dt_ordinal', 'dt_YYYYMMDD', 'lat', 'lon', 'z_WGS84']

    """
    ICESat-1 filters
    """
    #Saturation Correction Flag
    #These are 0 to 5, not_saturated inconsequential applicable not_computed not_applicable
    sat_corr_flg = f.get('Data_40HZ/Quality/sat_corr_flg')[mask]
    #valid_idx *= (sat_corr_flg < 2)

    #Correction to elevation for saturated waveforms
    #Notes suggest this might not be desirable over land
    satElevCorr = np.ma.masked_equal(f.get('Data_40HZ/Elevation_Corrections/d_satElevCorr')[mask], 1.7976931348623157e+308)
    #z[sat_corr_flg < 3] += satElevCorr.filled(0.0)[sat_corr_flg < 3]
    out[:,zcol] += satElevCorr.filled(0.0)

    #Correction to elevation based on post flight analysis for biases determined for each campaign
    ElevBiasCorr = np.ma.masked_equal(f.get('Data_40HZ/Elevation_Corrections/d_ElevBiasCorr')[mask], 1.7976931348623157e+308)
    out[:,zcol] += ElevBiasCorr.filled(0.0)

    #Surface elevation (T/P ellipsoid) minus surface elevation (WGS84 ellipsoid).
    #Approximately 0.7 m, so WGS is lower; need to subtract from d_elev
    deltaEllip = np.ma.masked_equal(f.get('Data_40HZ/Geophysical/d_deltaEllip')[mask], 1.7976931348623157e+308)
    out[:,zcol] -= deltaEllip

    #These are 1 for valid, 0 for invalid
    valid_idx *= ~(np.ma.getmaskarray(out[:,zcol]))
    print("z corrections: %i" % valid_idx.nonzero()[0].size)

    if False:
        #Reflectivity, not corrected for atmospheric effects
        reflctUC = np.ma.masked_equal(f.get('Data_40HZ/Reflectivity/d_reflctUC')[mask], 1.7976931348623157e+308)
        #This was minimum used for ice sheets
        min_reflctUC = 0.025
        valid_idx *= (reflctUC > min_reflctUC).data
        print("reflctUC: %i" % valid_idx.nonzero()[0].size)

    if False:
        #The Standard deviation of the difference between the functional fit and the received echo \
        #using alternate parameters. It is directly taken from GLA05 parameter d_wfFitSDev_1
        LandVar = np.ma.masked_equal(f.get('Data_40HZ/Elevation_Surfaces/d_LandVar')[mask], 1.7976931348623157e+308)
        #This was max used for ice sheets
        max_LandVar = 0.04
        valid_idx *= (LandVar < max_LandVar).data
        print("LandVar: %i" % valid_idx.nonzero()[0].size)

    if True:
        #Flag indicating whether the elevations on this record should be used.
        #0 = valid, 1 = not valid
        elev_use_flg = f.get('Data_40HZ/Quality/elev_use_flg')[mask].astype('Bool')
        valid_idx *= ~elev_use_flg
        print("elev_use_flg: %i" % valid_idx.nonzero()[0].size)

    if False:
        #Cloud contamination; Indicates if Gain > flag value, indicating probable cloud contamination.
        elv_cloud_flg = f.get('Data_40HZ/Elevation_Flags/elv_cloud_flg')[mask].astype('Bool')
        valid_idx *= ~elv_cloud_flg
        print("elv_cloud_flg: %i" % valid_idx.nonzero()[0].size)

    if False: 
        #Full resolution 1064 Quality Flag; 0 - 12 indicate Cloud detected
        FRir_qa_flg = f.get('Data_40HZ/Atmosphere/FRir_qa_flg')[mask]
        valid_idx *= (FRir_qa_flg == 15).data
        print("FRir_qa_flg: %i" % valid_idx.nonzero()[0].size)

    if False:
        #This is elevation extracted from SRTM30
        DEM_elv = np.ma.masked_equal(f.get('Data_40HZ/Geophysical/d_DEM_elv')[mask], 1.7976931348623157e+308)
        z_DEM_diff = np.abs(out[:,zcol] - DEM_elv)
        valid_idx *= (z_DEM_diff < max_z_DEM_diff).data
        print("z_DEM_diff: %i" % valid_idx.nonzero()[0].size)

        #d_DEMhiresArElv is a 9 element array of high resolution DEM values. The array index corresponds to the position of the DEM value relative to the spot. (5) is the footprint center.
        DEMhiresArElv = np.ma.masked_equal(f.get('Data_40HZ/Geophysical/d_DEMhiresArElv')[mask], 1.7976931348623157e+308)
        DEMhiresArElv_std = np.ma.std(DEMhiresArElv, axis=1)
        valid_idx *= (DEMhiresArElv_std < max_DEMhiresArElv_std).data
        print("max_DEMhiresArElv_std: %i" % valid_idx.nonzero()[0].size)
        #Compute slope

    #Apply cumulative filter to output
    out = out[valid_idx]

    out_fn = os.path.splitext(fn)[0]+'_%s.csv' % sitename
    print("Writing out %i records to: %s\n" % (out.shape[0], out_fn))
    out_fmt_str = ', '.join(out_fmt)
    out_hdr_str = ', '.join(out_hdr)
    np.savetxt(out_fn, out, fmt=out_fmt_str, delimiter=',', header=out_hdr_str)
    iolib.writevrt(out_fn, x='lon', y='lat')

    #Extract our own DEM values - should be better than default GLAS reference DEM stats
    if True:
        print("Loading reference DEM: %s" % refdem_fn)
        dem_ds = gdal.Open(refdem_fn)
        print("Converting coords for DEM")
        dem_mX, dem_mY = geolib.ds_cT(dem_ds, out[:,xcol], out[:,ycol], geolib.wgs_srs)
        print("Sampling")
        dem_samp = geolib.sample(dem_ds, dem_mX, dem_mY, pad='glas')
        abs_dem_z_diff = np.abs(out[:,zcol] - dem_samp[:,0])

        valid_idx *= ~(np.ma.getmaskarray(abs_dem_z_diff))
        print("Valid DEM extract: %i" % valid_idx.nonzero()[0].size)
        valid_idx *= (abs_dem_z_diff < max_z_DEM_diff).data
        print("Valid abs DEM diff: %i" % valid_idx.nonzero()[0].size)
        valid_idx *= (dem_samp[:,1] < max_DEMhiresArElv_std).data
        print("Valid DEM mad: %i" % valid_idx.nonzero()[0].size)

        if valid_idx.nonzero()[0].size == 0:
            sys.exit("No valid points remain")

        out = np.ma.hstack([out, dem_samp])
        out_fmt.extend(['%0.2f', '%0.2f'])
        out_hdr.extend(['z_refdem_med_WGS84', 'z_refdem_nmad'])

        #Apply cumulative filter to output
        out = out[valid_idx]

        out_fn = os.path.splitext(out_fn)[0]+'_refdemfilt.csv'
        print("Writing out %i records to: %s\n" % (out.shape[0], out_fn))
        out_fmt_str = ', '.join(out_fmt)
        out_hdr_str = ', '.join(out_hdr)
        np.savetxt(out_fn, out, fmt=out_fmt_str, delimiter=',', header=out_hdr_str)
        iolib.writevrt(out_fn, x='lon', y='lat')

    #This will sample land-use/land-cover or percent bareground products
    #Can be used to isolate points over exposed rock
    #if args.rockfilter: 
    if True:
        #This should automatically identify appropriate LULC source based on refdem extent
        lulc_source = dem_mask.get_lulc_source(dem_ds)
        #Looks like NED extends beyond NCLD, force use NLCD for conus
        #if sitename == 'conus':
        #    lulc_source = 'nlcd'
        lulc_ds = dem_mask.get_lulc_ds_full(dem_ds, lulc_source)
        print("Converting coords for LULC")
        lulc_mX, lulc_mY = geolib.ds_cT(lulc_ds, out[:,xcol], out[:,ycol], geolib.wgs_srs)
        print("Sampling LULC: %s" % lulc_source)
        #Note: want to make sure we're not interpolating integer values for NLCD
        #Should be safe with pad=0, even with pad>0, should take median, not mean
        lulc_samp = geolib.sample(lulc_ds, lulc_mX, lulc_mY, pad=0)
        l = lulc_samp[:,0].data
        if lulc_source == 'nlcd':
            #This passes rock and ice pixels
            valid_idx = np.logical_or((l==31),(l==12))
        elif lulc_source == 'bareground':
            #This preserves pixels with bareground percentation >85%
            minperc = 85
            valid_idx = (l >= minperc)
        else:
            print("Unknown LULC source")
        print("LULC: %i" % valid_idx.nonzero()[0].size)
        if l.ndim == 1:
            l = l[:,np.newaxis]
        out = np.ma.hstack([out, l])
        out_fmt.append('%i')
        out_hdr.append('lulc')

        #Apply cumulative filter to output
        out = out[valid_idx]

        out_fn = os.path.splitext(out_fn)[0]+'_lulcfilt.csv'
        print("Writing out %i records to: %s\n" % (out.shape[0], out_fn))
        out_fmt_str = ', '.join(out_fmt)
        out_hdr_str = ', '.join(out_hdr)
        np.savetxt(out_fn, out, fmt=out_fmt_str, delimiter=',', header=out_hdr_str)
        iolib.writevrt(out_fn, x='lon', y='lat')

if __name__ == "__main__":
    main()



#! /usr/bin/env python
"""
Utility to automate reference surface identification for raster co-registration

Note: Initial run may take a long time to download and process required data (NLCD, global bareground, glacier polygons)

Can control location of these data files with DATADIR environmental variable

export DATADIR=dir

Dependencies: gdal, wget, requests, bs4

"""

#To do: 
#Integrate 1-km LULC data: http://www.landcover.org/data/landcover/
#TODO: need to clean up toa handling

import sys
import os
import subprocess
import glob
import argparse

from osgeo import gdal, ogr, osr
import numpy as np

from datetime import datetime, timedelta

from pygeotools.lib import iolib, warplib, geolib, timelib

datadir = iolib.get_datadir()

def get_nlcd_fn():
    """Calls external shell script `get_nlcd.sh` to fetch:

    2011 Land Use Land Cover (nlcd) grids, 30 m
    
    http://www.mrlc.gov/nlcd11_leg.php
    """
    #This is original filename, which requires ~17 GB
    #nlcd_fn = os.path.join(datadir, 'nlcd_2011_landcover_2011_edition_2014_10_10/nlcd_2011_landcover_2011_edition_2014_10_10.img')
    #get_nlcd.sh now creates a compressed GTiff, which is 1.1 GB
    nlcd_fn = os.path.join(datadir, 'nlcd_2011_landcover_2011_edition_2014_10_10/nlcd_2011_landcover_2011_edition_2014_10_10.tif')
    if not os.path.exists(nlcd_fn):
        cmd = ['get_nlcd.sh',]
        #subprocess.call(cmd)
        sys.exit("Missing nlcd data source. If already downloaded, specify correct datadir. If not, run `%s` to download" % cmd[0])
    return nlcd_fn

def get_bareground_fn():
    """Calls external shell script `get_bareground.sh` to fetch:

    ~2010 global bare ground, 30 m

    Note: unzipped file size is 64 GB! Original products are uncompressed, and tiles are available globally (including empty data over ocean)

    The shell script will compress all downloaded tiles using lossless LZW compression.

    http://landcover.usgs.gov/glc/BareGroundDescriptionAndDownloads.php
    """
    bg_fn = os.path.join(datadir, 'bare2010/bare2010.vrt')
    if not os.path.exists(bg_fn):
        cmd = ['get_bareground.sh',]
        sys.exit("Missing bareground data source. If already downloaded, specify correct datadir. If not, run `%s` to download" % cmd[0])
        #subprocess.call(cmd)
    return bg_fn 

#Download latest global RGI glacier db
def get_glacier_poly():
    """Calls external shell script `get_rgi.sh` to fetch:

    Randolph Glacier Inventory (RGI) glacier outline shapefiles 

    Full RGI database: rgi50.zip is 410 MB

    The shell script will unzip and merge regional shp into single global shp
    
    http://www.glims.org/RGI/
    """
    #rgi_fn = os.path.join(datadir, 'rgi50/regions/rgi50_merge.shp')
    #Update to rgi60, should have this returned from get_rgi.sh
    rgi_fn = os.path.join(datadir, 'rgi60/regions/rgi60_merge.shp')
    if not os.path.exists(rgi_fn):
        cmd = ['get_rgi.sh',]
        sys.exit("Missing rgi glacier data source. If already downloaded, specify correct datadir. If not, run `%s` to download" % cmd[0])
        #subprocess.call(cmd)
    return rgi_fn 

#Update glacier polygons
def get_icemask(ds, glac_shp_fn=None):
    """Generate glacier polygon raster mask for input Dataset res/extent
    """
    print("Masking glaciers")
    if glac_shp_fn is None:
        glac_shp_fn = get_glacier_poly()

    if not os.path.exists(glac_shp_fn):
        print("Unable to locate glacier shp: %s" % glac_shp_fn)
    else:
        print("Found glacier shp: %s" % glac_shp_fn)

    #All of the proj, extent, handling should now occur in shp2array
    icemask = geolib.shp2array(glac_shp_fn, ds)
    return icemask

#Create nlcd mask 
def get_nlcd_mask(nlcd_ds, filter='not_forest', out_fn=None):
    """Generate raster mask for specified NLCD LULC filter
    """
    print("Loading NLCD LULC")
    b = nlcd_ds.GetRasterBand(1)
    l = b.ReadAsArray()
    print("Filtering NLCD LULC with: %s" % filter)
    #Original nlcd products have nan as ndv
        #12 - ice
        #31 - rock
        #11 - open water, includes rivers
        #52 - shrub, <5 m tall, >20%
        #42 - evergreeen forest
    #Should use data dictionary here for general masking
    #Using 'rock+ice+water' preserves the most pixels, although could be problematic over areas with lakes
    if filter == 'rock':
        mask = (l==31)
    elif filter == 'rock+ice':
        mask = np.logical_or((l==31),(l==12))
    elif filter == 'rock+ice+water':
        mask = np.logical_or(np.logical_or((l==31),(l==12)),(l==11))
    elif filter == 'not_forest':
        mask = ~(np.logical_or(np.logical_or((l==41),(l==42)),(l==43)))
    elif filter == 'not_forest+not_water':
        mask = ~(np.logical_or(np.logical_or(np.logical_or((l==41),(l==42)),(l==43)),(l==11)))
    else:
        print("Invalid mask type")
        mask = None
    #Write out original data
    if out_fn is not None:
        print("Writing out %s" % out_fn)
        iolib.writeGTiff(l, out_fn, nlcd_ds)
    l = None
    return mask

def get_bareground_mask(bareground_ds, bareground_thresh=60, out_fn=None):
    """Generate raster mask for exposed bare ground from global bareground data
    """
    print("Loading bareground")
    b = bareground_ds.GetRasterBand(1)
    l = b.ReadAsArray()
    print("Masking pixels with <%0.1f%% bare ground" % bareground_thresh)
    if bareground_thresh < 0.0 or bareground_thresh > 100.0:
        sys.exit("Invalid bare ground percentage")
    mask = (l>bareground_thresh)
    #Write out original data
    if out_fn is not None:
        print("Writing out %s" % out_fn)
        iolib.writeGTiff(l, out_fn, bareground_ds)
    l = None
    return mask

def get_snodas_ds(dem_dt, code=1036):
    """Function to fetch and process SNODAS snow depth products for input datetime

    http://nsidc.org/data/docs/noaa/g02158_snodas_snow_cover_model/index.html

    Product codes:
    1036 is snow depth
    1034 is SWE

    filename format: us_ssmv11036tS__T0001TTNATS2015042205HP001.Hdr

    """
    import tarfile
    import gzip
    snodas_ds = None
    snodas_url_str = None

    outdir = os.path.join(datadir, 'snodas')
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    #Note: unmasked products (beyond CONUS) are only available from 2010-present
    if dem_dt >= datetime(2003,9,30) and dem_dt < datetime(2010,1,1):
        snodas_url_str = 'ftp://sidads.colorado.edu/DATASETS/NOAA/G02158/masked/%Y/%m_%b/SNODAS_%Y%m%d.tar'
        tar_subfn_str_fmt = 'us_ssmv1%itS__T0001TTNATS%%Y%%m%%d05HP001.%s.gz'
    elif dem_dt >= datetime(2010,1,1):
        snodas_url_str = 'ftp://sidads.colorado.edu/DATASETS/NOAA/G02158/unmasked/%Y/%m_%b/SNODAS_unmasked_%Y%m%d.tar'
        tar_subfn_str_fmt = './zz_ssmv1%itS__T0001TTNATS%%Y%%m%%d05HP001.%s.gz'
    else:
        print("No SNODAS data available for input date")

    if snodas_url_str is not None:
        snodas_url = dem_dt.strftime(snodas_url_str)
        snodas_tar_fn = iolib.getfile(snodas_url, outdir=outdir)
        print("Unpacking")
        tar = tarfile.open(snodas_tar_fn)
        #gunzip to extract both dat and Hdr files, tar.gz
        for ext in ('dat', 'Hdr'):
            tar_subfn_str = tar_subfn_str_fmt % (code, ext)
            tar_subfn_gz = dem_dt.strftime(tar_subfn_str)
            tar_subfn = os.path.splitext(tar_subfn_gz)[0]
            print(tar_subfn)
            if outdir is not None:
                tar_subfn = os.path.join(outdir, tar_subfn)
            if not os.path.exists(tar_subfn):
                #Should be able to do this without writing intermediate gz to disk
                tar.extract(tar_subfn_gz)
                with gzip.open(tar_subfn_gz, 'rb') as f:
                    outf = open(tar_subfn, 'wb')
                    outf.write(f.read())
                    outf.close()
                os.remove(tar_subfn_gz)

        #Need to delete 'Created by module comment' line from Hdr, can contain too many characters
        bad_str = 'Created by module comment'
        snodas_fn = tar_subfn
        f = open(snodas_fn)
        output = []
        for line in f:
            if not bad_str in line:
                output.append(line)
        f.close()
        f = open(snodas_fn, 'w')
        f.writelines(output)
        f.close()

        #Return GDAL dataset for extracted product
        snodas_ds = gdal.Open(snodas_fn)
    return snodas_ds

def get_modis_tile_list(ds):
    """Helper function to identify MODIS tiles that intersect input geometry

    modis_gird.py contains dictionary of tile boundaries (tile name and WKT polygon ring from bbox)

    See: https://modis-land.gsfc.nasa.gov/MODLAND_grid.html
    """
    from demcoreg import modis_grid
    modis_dict = {}
    for key in modis_grid.modis_dict:
        modis_dict[key] = ogr.CreateGeometryFromWkt(modis_grid.modis_dict[key])
    geom = geolib.ds_geom(ds)
    geom_dup = geolib.geom_dup(geom)
    ct = osr.CoordinateTransformation(geom_dup.GetSpatialReference(), geolib.wgs_srs)
    geom_dup.Transform(ct)
    tile_list = []
    for key, val in list(modis_dict.items()):
        if geom_dup.Intersects(val):
            tile_list.append(key)
    return tile_list

def get_modscag_fn_list(dem_dt, tile_list=('h08v04', 'h09v04', 'h10v04', 'h08v05', 'h09v05'), pad_days=7):
    """Function to fetch and process MODSCAG fractional snow cover products for input datetime

    Products are tiled in MODIS sinusoidal projection

    example url: https://snow-data.jpl.nasa.gov/modscag-historic/2015/001/MOD09GA.A2015001.h07v03.005.2015006001833.snow_fraction.tif

    """

    #Could also use global MODIS 500 m snowcover grids, 8 day
    #http://nsidc.org/data/docs/daac/modis_v5/mod10a2_modis_terra_snow_8-day_global_500m_grid.gd.html
    #These are HDF4, sinusoidal
    #Should be able to load up with warplib without issue

    import re
    import requests
    from bs4 import BeautifulSoup
    auth = iolib.get_auth()
    pad_days = timedelta(days=pad_days)
    dt_list = timelib.dt_range(dem_dt-pad_days, dem_dt+pad_days+timedelta(1), timedelta(1))

    outdir = os.path.join(datadir, 'modscag')
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    out_vrt_fn_list = []
    for dt in dt_list:
        out_vrt_fn = os.path.join(outdir, dt.strftime('%Y%m%d_snow_fraction.vrt'))
        #If we already have a vrt and it contains all of the necessary tiles
        if os.path.exists(out_vrt_fn):
            vrt_ds = gdal.Open(out_vrt_fn)
            if np.all([np.any([tile in sub_fn for sub_fn in vrt_ds.GetFileList()]) for tile in tile_list]):
                out_vrt_fn_list.append(out_vrt_fn)
                continue
        #Otherwise, download missing tiles and rebuild
        #Try to use historic products
        modscag_fn_list = []
        #Note: not all tiles are available for same date ranges in historic vs. real-time
        #Need to repeat search tile-by-tile
        for tile in tile_list:
            modscag_url_str = 'https://snow-data.jpl.nasa.gov/modscag-historic/%Y/%j/' 
            modscag_url_base = dt.strftime(modscag_url_str)
            print("Trying: %s" % modscag_url_base)
            r = requests.get(modscag_url_base, auth=auth)
            modscag_url_fn = []
            if r.ok:
                parsed_html = BeautifulSoup(r.content, "html.parser")
                modscag_url_fn = parsed_html.findAll(text=re.compile('%s.*snow_fraction.tif' % tile))
            if not modscag_url_fn:
                #Couldn't find historic, try to use real-time products
                modscag_url_str = 'https://snow-data.jpl.nasa.gov/modscag/%Y/%j/' 
                modscag_url_base = dt.strftime(modscag_url_str)
                print("Trying: %s" % modscag_url_base)
                r = requests.get(modscag_url_base, auth=auth)
            if r.ok: 
                parsed_html = BeautifulSoup(r.content, "html.parser")
                modscag_url_fn = parsed_html.findAll(text=re.compile('%s.*snow_fraction.tif' % tile))
            if not modscag_url_fn:
                print("Unable to fetch MODSCAG for %s" % dt)
            else:
                #OK, we got
                #Now extract actual tif filenames to fetch from html
                parsed_html = BeautifulSoup(r.content, "html.parser")
                #Fetch all tiles
                modscag_url_fn = parsed_html.findAll(text=re.compile('%s.*snow_fraction.tif' % tile))
                if modscag_url_fn:
                    modscag_url_fn = modscag_url_fn[0]
                    modscag_url = os.path.join(modscag_url_base, modscag_url_fn)
                    print(modscag_url)
                    modscag_fn = os.path.join(outdir, os.path.split(modscag_url_fn)[-1])
                    if not os.path.exists(modscag_fn):
                        iolib.getfile2(modscag_url, auth=auth, outdir=outdir)
                    modscag_fn_list.append(modscag_fn)
        #Mosaic tiles - currently a hack
        if modscag_fn_list:
            cmd = ['gdalbuildvrt', '-vrtnodata', '255', out_vrt_fn]
            cmd.extend(modscag_fn_list)
            print(cmd)
            subprocess.call(cmd, shell=False)
            out_vrt_fn_list.append(out_vrt_fn)
    return out_vrt_fn_list

def proc_modscag(fn_list, extent=None, t_srs=None):
    """Process the MODSCAG products for full date range, create composites and reproject
    """
    #Use cubic spline here for improve upsampling 
    ds_list = warplib.memwarp_multi_fn(fn_list, res='min', extent=extent, t_srs=t_srs, r='cubicspline')
    stack_fn = os.path.splitext(fn_list[0])[0] + '_' + os.path.splitext(os.path.split(fn_list[-1])[1])[0] + '_stack_%i' % len(fn_list) 
    #Create stack here - no need for most of mastack machinery, just make 3D array
    #Mask values greater than 100% (clouds, bad pixels, etc)
    ma_stack = np.ma.array([np.ma.masked_greater(iolib.ds_getma(ds), 100) for ds in np.array(ds_list)], dtype=np.uint8)

    stack_count = np.ma.masked_equal(ma_stack.count(axis=0), 0).astype(np.uint8)
    stack_count.set_fill_value(0)
    stack_min = ma_stack.min(axis=0).astype(np.uint8)
    stack_min.set_fill_value(0)
    stack_max = ma_stack.max(axis=0).astype(np.uint8)
    stack_max.set_fill_value(0)
    stack_med = np.ma.median(ma_stack, axis=0).astype(np.uint8)
    stack_med.set_fill_value(0)

    out_fn = stack_fn + '_count.tif'
    iolib.writeGTiff(stack_count, out_fn, ds_list[0])
    out_fn = stack_fn + '_max.tif'
    iolib.writeGTiff(stack_max, out_fn, ds_list[0])
    out_fn = stack_fn + '_min.tif'
    iolib.writeGTiff(stack_min, out_fn, ds_list[0])
    out_fn = stack_fn + '_med.tif'
    iolib.writeGTiff(stack_med, out_fn, ds_list[0])

    ds = gdal.Open(out_fn)
    return ds

def get_toa_fn(dem_fn):
    toa_fn = None
    #Original approach, assumes DEM file is in *00/dem_*/*DEM_32m.tif
    #dem_dir = os.path.split(os.path.split(os.path.abspath(dem_fn))[0])[0]
    dem_dir_list = os.path.split(os.path.realpath(dem_fn))[0].split(os.sep)
    import re
    #Get index of the top level pair directory containing toa (WV02_20140514_1030010031114100_1030010030896000)
    r_idx = [i for i, item in enumerate(dem_dir_list) if re.search('(_10)*(_10)*00$', item)]
    if r_idx:
        r_idx = r_idx[0]
        #Reconstruct dir
        dem_dir = (os.sep).join(dem_dir_list[0:r_idx+1])
        #Find toa.tif in top-level dir
        toa_fn = glob.glob(os.path.join(dem_dir, '*toa.tif'))
        if not toa_fn:
            ortho_fn = glob.glob(os.path.join(dem_dir, '*ortho*.tif'))
            if ortho_fn:
                cmd = ['toa.sh', dem_dir]
                print(cmd)
                subprocess.call(cmd)
                toa_fn = glob.glob(os.path.join(dem_dir, '*toa.tif'))
        if toa_fn:
            toa_fn = toa_fn[0]
        else:
            toa_fn = None
    if toa_fn is None:
        sys.exit("Unable to locate TOA dataset")
    return toa_fn

#TOA reflectance filter
def get_toa_mask(toa_ds, toa_thresh=0.4):
    print("Applying TOA filter (masking values >= %0.2f)" % toa_thresh)
    toa = iolib.ds_getma(toa_ds)
    toa_mask = np.ma.masked_greater(toa, toa_thresh)
    #This should be 1 for valid surfaces, 0 for snowcovered surfaces
    toa_mask = ~(np.ma.getmaskarray(toa_mask))
    return toa_mask

def check_mask_list(mask_list):
    temp = []
    for m in mask_list:
        if m not in mask_choices: 
            print("Invalid mask choice: %s" % m)
        else:
            temp.append(m)
    return temp

def get_mask(dem_ds, mask_list, dem_fn=None, writeout=False, outdir=None, args=None):
    mask_list = check_mask_list(mask_list)
    if 'none' in mask_list:
        newmask = False
    else:
        #Basename for output files
        if outdir is not None:
            if not os.path.exists(outdir):
                os.makedirs(outdir)
        else:
            outdir = os.path.split(dem_fn)[0]

        if dem_fn is not None:
            #Extract DEM timestamp
            dem_dt = timelib.fn_getdatetime(dem_fn)
            out_fn_base = os.path.join(outdir, os.path.splitext(dem_fn)[0])
        
        if args is None:
            #Get default values
            parser = getparser()
            args = parser.parse_args(['',])

        newmask = True
        
        if 'glaciers' in mask_list:
            icemask = get_icemask(dem_ds)
            if writeout:
                out_fn = out_fn_base+'_ice_mask.tif'
                print("Writing out %s" % out_fn)
                iolib.writeGTiff(icemask, out_fn, src_ds=dem_ds)
            newmask = np.logical_and(icemask, newmask)

        #Need to process NLCD separately, with nearest neighbor inteprolatin
        if 'nlcd' in mask_list and args.nlcd_filter is not 'none':
            rs = 'near'
            nlcd_ds = gdal.Open(get_nlcd_fn())
            nlcd_ds_warp = warplib.memwarp_multi([nlcd_ds,], res=dem_ds, extent=dem_ds, t_srs=dem_ds, r=rs)[0]
            out_fn = None
            if writeout:
                out_fn = out_fn_base+'_nlcd.tif'
            nlcdmask = get_nlcd_mask(nlcd_ds_warp, filter=args.nlcd_filter, out_fn=out_fn)
            if writeout:
                out_fn = os.path.splitext(out_fn)[0]+'_mask.tif'
                print("Writing out %s" % out_fn)
                iolib.writeGTiff(nlcdmask, out_fn, src_ds=dem_ds)
            newmask = np.logical_and(nlcdmask, newmask)

        if 'bareground' in mask_list and args.bareground_thresh > 0:
            bareground_ds = gdal.Open(get_bareground_fn())
            bareground_ds_warp = warplib.memwarp_multi([bareground_ds,], res=dem_ds, extent=dem_ds, t_srs=dem_ds, r='cubicspline')[0]
            out_fn = None
            if writeout:
                out_fn = out_fn_base+'_bareground.tif'
            baregroundmask = get_bareground_mask(bareground_ds_warp, bareground_thresh=args.bareground_thresh, out_fn=out_fn)
            if writeout:
                out_fn = os.path.splitext(out_fn)[0]+'_mask.tif'
                print("Writing out %s" % out_fn)
                iolib.writeGTiff(baregroundmask, out_fn, src_ds=dem_ds)
            newmask = np.logical_and(baregroundmask, newmask)

        if 'snodas' in mask_list and args.snodas_thresh > 0:
            #Get SNODAS snow depth products for DEM timestamp
            snodas_min_dt = datetime(2003,9,30)
            if dem_dt >= snodas_min_dt: 
                snodas_ds = get_snodas_ds(dem_dt)
                if snodas_ds is not None:
                    snodas_ds_warp = warplib.memwarp_multi([snodas_ds,], res=dem_ds, extent=dem_ds, t_srs=dem_ds, r='cubicspline')[0]
                    #snow depth values are mm, convert to meters
                    snodas_depth = iolib.ds_getma(snodas_ds_warp)/1000.
                    if snodas_depth.count() > 0:
                        print("Applying SNODAS snow depth filter (masking values >= %0.2f m)" % args.snodas_thresh)
                        out_fn = None
                        if writeout:
                            out_fn = out_fn_base+'_snodas_depth.tif'
                            print("Writing out %s" % out_fn)
                            iolib.writeGTiff(snodas_depth, out_fn, src_ds=dem_ds)
                        snodas_mask = np.ma.masked_greater(snodas_depth, args.snodas_thresh)
                        snodas_mask = ~(np.ma.getmaskarray(snodas_mask))
                        if writeout:
                            out_fn = os.path.splitext(out_fn)[0]+'_mask.tif'
                            print("Writing out %s" % out_fn)
                            iolib.writeGTiff(snodas_mask, out_fn, src_ds=dem_ds)
                        newmask = np.logical_and(snodas_mask, newmask)
                    else:
                        print("SNODAS grid for input location and timestamp is empty")

        #These tiles cover CONUS
        #tile_list=('h08v04', 'h09v04', 'h10v04', 'h08v05', 'h09v05')
        if 'modscag' in mask_list and args.modscag_thresh > 0:
            modscag_min_dt = datetime(2000,2,24)
            if dem_dt < modscag_min_dt: 
                print("Warning: DEM timestamp (%s) is before earliest MODSCAG timestamp (%s)" \
                        % (dem_dt, modscag_min_dt))
            else:
                tile_list = get_modis_tile_list(dem_ds)
                print(tile_list)
                pad_days=7
                modscag_fn_list = get_modscag_fn_list(dem_dt, tile_list=tile_list, pad_days=pad_days)
                if modscag_fn_list:
                    modscag_ds = proc_modscag(modscag_fn_list, extent=dem_ds, t_srs=dem_ds)
                    modscag_ds_warp = warplib.memwarp_multi([modscag_ds,], res=dem_ds, extent=dem_ds, t_srs=dem_ds, r='cubicspline')[0]
                    print("Applying MODSCAG fractional snow cover percent filter (masking values >= %0.1f%%)" % args.modscag_thresh)
                    modscag_fsca = iolib.ds_getma(modscag_ds_warp)
                    out_fn = None
                    if writeout:
                        out_fn = out_fn_base+'_modscag_fsca.tif'
                        print("Writing out %s" % out_fn)
                        iolib.writeGTiff(modscag_fsca, out_fn, src_ds=dem_ds)
                    modscag_mask = (modscag_fsca.filled(0) >= args.modscag_thresh) 
                    modscag_mask = ~(modscag_mask)
                    if writeout:
                        out_fn = os.path.splitext(out_fn)[0]+'_mask.tif'
                        print("Writing out %s" % out_fn)
                        iolib.writeGTiff(modscag_mask, out_fn, src_ds=dem_ds)
                    newmask = np.logical_and(modscag_mask, newmask)

        #Use reflectance values to estimate snowcover
        if 'toa' in mask_list:
            #Use top of atmosphere scaled reflectance values (0-1)
            toa_ds = gdal.Open(get_toa_fn(dem_fn))
            toa_mask = get_toa_mask(toa_ds, args.toa_thresh)
            if writeout:
                out_fn = out_fn_base+'_toa_mask.tif'
                print("Writing out %s" % out_fn)
                iolib.writeGTiff(toa_mask, out_fn, src_ds=dem_ds)
            newmask = np.logical_and(toa_mask, newmask)

        if False:
            #Filter based on expected snowline
            #Simplest approach uses altitude cutoff
            max_elev = 1500 
            newdem = np.ma.masked_greater(dem, max_elev)
            newmask = np.ma.getmaskarray(newdem)

        print("Generating final mask to use for reference surfaces, and applying to input DEM")
        #Now invert to use to create final masked array
        #True (1) represents "invalid" pixel to match numpy ma convetion 
        newmask = ~newmask

        #Dilate the mask
        if args.dilate is not None:
            niter = args.dilate 
            print("Dilating mask with %i iterations" % niter)
            from scipy import ndimage
            newmask = ~(ndimage.morphology.binary_dilation(~newmask, iterations=niter))

    return newmask

#Can add "mask_list" argument, instead of specifying individually
mask_choices = ['toa', 'snodas', 'modscag', 'bareground', 'glaciers', 'nlcd', 'none']
def getparser():
    parser = argparse.ArgumentParser(description="Identify control surfaces for DEM co-registration") 
    parser.add_argument('dem_fn', type=str, help='DEM filename')
    parser.add_argument('--outdir', default=None, help='Directory for output products')
    parser.add_argument('--writeout', action='store_true', help='Write out all intermediate products, instead of only final tif')
    #parser.add_argument('-datadir', default=None, help='Data directory containing reference data sources (NLCD, bareground, etc)')
    parser.add_argument('--toa', action='store_true', help='Use top-of-atmosphere reflectance values (requires pregenerated "dem_fn_toa.tif")')
    parser.add_argument('--toa_thresh', type=float, default=0.4, help='Top-of-atmosphere reflectance threshold (default: %(default)s, valid range 0.0-1.0), mask values greater than this value')
    parser.add_argument('--snodas', action='store_true', help='Use SNODAS snow depth products')
    parser.add_argument('--snodas_thresh', type=float, default=0.2, help='SNODAS snow depth threshold (default: %(default)s m), mask values greater than this value')
    parser.add_argument('--modscag', action='store_true', help='Use MODSCAG fractional snow cover products')
    parser.add_argument('--modscag_thresh', type=float, default=50, help='MODSCAG fractional snow cover percent threshold (default: %(default)s%%, valid range 0-100), mask greater than this value')
    parser.add_argument('--bareground', action='store_true', help="Enable bareground filter")
    parser.add_argument('--bareground_thresh', type=float, default=60, help='Percent bareground threshold (default: %(default)s%%, valid range 0-100), mask greater than this value (only relevant for global bareground data)')
    parser.add_argument('--glaciers', action='store_true', help="Mask glacier polygons")
    parser.add_argument('--nlcd', action='store_true', help="Enable NLCD LULC filter (for CONUS)")
    nlcd_filter_choices = ['rock', 'rock+ice', 'rock+ice+water', 'not_forest', 'not_forest+not_water', 'none']
    parser.add_argument('--nlcd_filter', type=str, default='not_forest', choices=nlcd_filter_choices, help='Preserve these NLCD pixels (default: %(default)s)') 
    parser.add_argument('--dilate', type=int, default=None, help='Dilate mask with this many iterations (default: %(default)s)')
    return parser

def main():
    parser = getparser()
    args = parser.parse_args()

    mask_list = []
    if args.toa: mask_list.append('toa') 
    if args.snodas: mask_list.append('snodas') 
    if args.modscag: mask_list.append('modscag') 
    if args.bareground: mask_list.append('bareground') 
    if args.glaciers: mask_list.append('glaciers') 
    if args.nlcd: mask_list.append('nlcd') 

    if not mask_list:
        parser.print_help()
        sys.exit("Must specify at least one mask type")

    #This directory should or will contain the relevant data products
    #if args.datadir is None:
    #    datadir = iolib.get_datadir() 

    dem_fn = args.dem_fn
    dem_ds = gdal.Open(dem_fn)
    print(dem_fn)

    #Get DEM masked array
    dem = iolib.ds_getma(dem_ds)
    print("%i valid pixels in original input tif" % dem.count())

    #Set up cascading mask preparation
    #True (1) represents "valid" unmasked pixel, False (0) represents "invalid" pixel to be masked
    #Initialize the mask
    #newmask = ~(np.ma.getmaskarray(dem))
    
    newmask = get_mask(dem_ds, mask_list, dem_fn=dem_fn, writeout=args.writeout, outdir=args.outdir, args=args)
    
    #Apply mask to original DEM - use these surfaces for co-registration
    newdem = np.ma.array(dem, mask=newmask)

    #Check that we have enough pixels, good distribution
    min_validpx_count = 100
    min_validpx_std = 10
    validpx_count = newdem.count()
    validpx_std = newdem.std()
    print("%i valid pixels in masked output tif to be used as ref" % validpx_count)
    print("%0.2f std in masked output tif to be used as ref" % validpx_std)
    #if (validpx_count > min_validpx_count) and (validpx_std > min_validpx_std):
    if (validpx_count > min_validpx_count):
        out_fn = os.path.join(args.outdir, os.path.splitext(dem_fn)[0]+'_ref.tif')
        print("Writing out %s" % out_fn)
        iolib.writeGTiff(newdem, out_fn, src_ds=dem_ds)
    else:
        print("Not enough valid pixels!")


if __name__ == "__main__":
    main()
